package org.netkernel.module.standard;

import java.util.*;
import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.meta.IPrototypeMeta;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.util.FastSchematron;
import org.netkernel.layer0.util.PairList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.builtin.PrivateFilterEndpoint;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.ISpace;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class SpaceFactory
{
	private static final PairList ROOTSPACE_VALIDATION = new PairList(10);
	static
	{	ROOTSPACE_VALIDATION.put("./attribute::node()[name()!='public' and name()!='uri' and name()!='name' and name()!='private-filter']", "MSG_STD_SPACE_1");
		ROOTSPACE_VALIDATION.put("/@public[.!='true' and .!='false']", "MSG_STD_SPACE_2");
		ROOTSPACE_VALIDATION.put("@private-filter[.!='true' and .!='false']", "MSG_STD_SPACE_3");
	}
	
	public static void buildRootSpace(RootSpace aRootSpace, Element aConfig, IKernel aKernel) throws Exception
	{
		//validate
		PairList validation=FastSchematron.validate(aConfig, ROOTSPACE_VALIDATION);
		if (validation.size()>0)
		{	String m = FastSchematron.formatValidationError(validation,aKernel.getLogger());			
			throw Utils.createFormattedException("EX_STD_CONSTRUCT", "MSG_RAW", aKernel.getLogger(), m);
		}
		
		//determine if we need a private filter and 
		XMLReadable rootReadable = new XMLReadable(aConfig);
		boolean hasPrivates=rootReadable.getFirstNode("*//private")!=null;
		boolean isFiltered=hasPrivates;
		String attValue=aConfig.getAttribute("private-filter");
		if (attValue.equals("true"))
		{	isFiltered=true;
		}
		else if (attValue.equals("false"))
		{	isFiltered=false;
		}
		
		if (isFiltered)
		{	StandardSpace insideSpace = new StandardSpace(aRootSpace,aRootSpace.getOwningModule(),aConfig);
			PrivateFilterEndpoint filter = new PrivateFilterEndpoint(insideSpace);
			aRootSpace.setMappings(new IStandardEndpoint[] {filter} );
			innerBuildSpace(insideSpace,aConfig,aKernel);
			String spaceName=insideSpace.getName();
			aRootSpace.setName(spaceName+" (public)");
		}
		else
		{	innerBuildSpace(aRootSpace,aConfig,aKernel);
		}
	}
	
	public static ISpace buildSubSpace(Element aConfig, StandardSpace aParent) throws Exception
	{	StandardSpace space = new StandardSpace(aParent,aParent.getOwningModule(),aConfig);
		IKernel k=aParent.getKernel();
		innerBuildSpace(space, aConfig, k);
		space.onCommissionSpace(k);
		return space;
	}
	
	public static void innerBuildSpace(StandardSpace aSpace, Element aConfig, IKernel aKernel) throws Exception
	{
		//determine name of space
		String spaceName=aConfig.getAttribute("name");
		if (spaceName.length()==0)
		{	String path=XMLUtils.getPathFor(aConfig);
			StandardSpace parent=aSpace.getParent();
			if (parent!=null)
			{	
				String parentPath=XMLUtils.getPathFor(parent.getConfigElement());
				String p2 = path.substring(parentPath.length());
				String pName=parent.getName();
				spaceName=pName+" "+p2;
			}
			else
			{	int i=path.indexOf('/', 1);
				spaceName=aSpace.getOwningModule().getMeta().getName()+" "+path.substring(i);
			}
		}
		aSpace.setName(spaceName);
		
		//build mappings
		List mappings = new ArrayList();
		for (Element n=XMLUtils.getFirstChildElement(aConfig); n!=null; n=XMLUtils.getNextSiblingElement(n) )
		{	
			XMLReadable r = new XMLReadable(n);
			String tagName=XMLUtils.pleaseDOMGiveMeTheNameOf(n);
			
			try
			{	IStandardEndpoint mapping=buildEndpoint(n, aSpace.getName(), aSpace.getClassLoader(), aKernel);
				mappings.add(mapping);
			}
			catch (INetKernelException e)
			{	logConstructionError(e,n,aSpace,aKernel);
				aSpace.setInitialisationProblem(Utils.throwableToString(e));
			}	
		}
		
		IStandardEndpoint mappingArray[] = new IStandardEndpoint[mappings.size()];
		mappings.toArray(mappingArray);
		aSpace.setMappings(mappingArray);
	}
	
	public static IStandardEndpoint buildEndpoint(Element aNode, String aLocation, ClassLoader aClassLoader, IKernel aKernel) throws INetKernelException
	{
		XMLReadable r = new XMLReadable(aNode);
		String tagName=XMLUtils.pleaseDOMGiveMeTheNameOf(aNode);
		IPrototypeMeta prototypeMeta=null;
		Class implementationClass;
		boolean excludeImplementationTags;
		boolean useTagAsParam=false;
		
		if (tagName.equals("overlay") || tagName.equals("endpoint") || tagName.equals("accessor") || tagName.equals("transreptor") || tagName.equals("transport"))
		{	String className=r.getTrimText("class");
			String prototypeName=r.getTrimText("prototype");
			excludeImplementationTags=true;
			if (className.length()==0 && prototypeName.length()==0)
			{	throw generateParseException("MSG_STD_UNDEFINED_OVERLAY_ENDPOINT",aLocation,XMLUtils.getPathFor(aNode),aKernel);
			}
			else if (className.length()==0)
			{	IPrototypeMeta builtinProto=lookupTag(prototypeName);
				if (builtinProto==null)
				{	implementationClass = LateBoundPrototype.class;
				}
				else
				{	implementationClass = builtinProto.getImplementation();	
				}
			}
			else if (prototypeName.length()==0)
			{	try
				{	implementationClass = aClassLoader.loadClass(className);
				}
				catch (ClassNotFoundException e)
				{	throw generateParseException("MSG_STD_EP_CLASS_NOT_FOUND",aLocation,e.getMessage(),aKernel);
				}
			}
			else
			{	throw generateParseException("MSG_STD_UNDEFINED_OVERLAY_ENDPOINT",aLocation,XMLUtils.getPathFor(aNode),aKernel);
			}
			/*
			if (tagName.equals("accessor"))
			{	if (!StandardAccessorImpl.class.isAssignableFrom(implementationClass))
				{	throw generateParseException("MSG_STD_NOT_ACCESSOR",XMLUtils.getPathFor(n),k);
				}
			}
			if (tagName.equals("transreptor"))
			{	if (!StandardTransreptorImpl.class.isAssignableFrom(implementationClass))
				{	throw generateParseException("MSG_STD_NOT_TRANSREPTOR",XMLUtils.getPathFor(n),k);
				}
			}
			*/
		}
		else 
		{	excludeImplementationTags=false;
			prototypeMeta = lookupTag(tagName);
			if (prototypeMeta==null)
			{	throw generateParseException("MSG_STD_UNDEFINED_TAG",aLocation,tagName,aKernel);
			}
			implementationClass=prototypeMeta.getImplementation();
			useTagAsParam = isTagSpace(tagName);
		}
		
		IStandardEndpoint mapping = constructEndpoint(implementationClass,aLocation,aKernel);
		if (prototypeMeta!=null)
		{	mapping.declareParameters(prototypeMeta.getParameters());
		}

		try
		{
			if (useTagAsParam)
			{	IPrototypeParameterMeta param1=prototypeMeta.getParameters()[0];
				mapping.setParameter(param1.getName(), aNode, aKernel);
			}
			else
			{	for (Element paramNode=XMLUtils.getFirstChildElement(aNode); paramNode!=null; paramNode=XMLUtils.getNextSiblingElement(paramNode) )
				{	String tagName2=XMLUtils.pleaseDOMGiveMeTheNameOf(paramNode);
					if (!(excludeImplementationTags && (tagName2.equals("class") || (tagName2.equals("prototype") && implementationClass!=LateBoundPrototype.class))))
					{	mapping.setParameter(tagName2, paramNode, aKernel);
					}
				}
			}
		}
		catch (Exception e)
		{	throw generateParseException("MSG_STD_EP_BAD_CONSTRUCTION",aLocation,null,aKernel,e);
		}
		return mapping;	
	}
	
	static IStandardEndpoint constructEndpoint(Class aImplementationClass, String aLocationMsg, IKernel aKernel) throws INetKernelException
	{	Object result;
		try
		{	result = aImplementationClass.newInstance();
		}
		catch (IllegalAccessException e)
		{	throw generateParseException("MSG_STD_EP_NO_CONSTR",aLocationMsg,e.getMessage(),aKernel);
		}
		catch (InstantiationException e)
		{	throw generateParseException("MSG_STD_EP_ABSTRACT",aLocationMsg,e.getMessage(),aKernel);
		}
		catch (Throwable t)
		{	throw generateParseException("MSG_STD_EP_BAD_CONSTRUCTION",aLocationMsg,null,aKernel,t);
		}	
		if (! (result instanceof IStandardEndpoint))
		{	throw generateParseException("MSG_STD_EP_NOT_COMPATIBLE",aLocationMsg,aImplementationClass.getName(),aKernel);
		}
		return (IStandardEndpoint)result;
	}
	
	static INetKernelException generateParseException(String aMessageId, String aLocation, Object aParam, IKernel aKernel)
	{	return generateParseException(aMessageId,aLocation,aParam,aKernel,null);
	}
	
	static INetKernelException generateParseException(String aMessageId, String aLocation, Object aParam, IKernel aKernel, Throwable aCause)
	{	return Utils.createFormattedException("EX_STD_CONSTRUCT", aMessageId,  aKernel.getLogger(), aCause, aLocation, aParam);
	}
	
	static void logConstructionError(INetKernelException e, Node aConfigElement, StandardSpace aSpace, IKernel aKernel)
	{	String owner=aSpace.getOwnerName();
		ILogger l = aKernel.getLogger();
		String path;
		if (aConfigElement!=null)
		{	path=XMLUtils.getPathFor(aConfigElement);
		}
		else
		{	path="";
		}
		String location=l.format("MSG_STD_CONSTRUCT_LOC",owner,path);
		NKFException e2 = new NKFException(e.getId(),e.getMessage(),location,e.getCause());
		String msg=e2.toString();
		l.logRaw(ILogger.WARNING, aSpace, msg);
	}		
	
	private static Map sTagToPrototypeMap = new HashMap();
	static
	{	sTagToPrototypeMap.put("representation", new org.netkernel.module.standard.builtin.MetaProviderPrototype());
		sTagToPrototypeMap.put("prototype", new org.netkernel.module.standard.builtin.MetaProviderPrototype());
		sTagToPrototypeMap.put("mapper", new org.netkernel.module.standard.builtin.mapper.MapperPrototype());
		sTagToPrototypeMap.put("fileset", new org.netkernel.module.standard.builtin.StaticResourcePrototype());
		sTagToPrototypeMap.put("import", new org.netkernel.module.standard.builtin.ImportPrototype());
		sTagToPrototypeMap.put("branch-merge", new org.netkernel.module.standard.builtin.branchmerge.BranchMergePrototype());
		sTagToPrototypeMap.put("pluggable-overlay", new org.netkernel.module.standard.builtin.PluggableOverlayPrototype());		
	}
	
	private static List<String> sTagsAsParam = new ArrayList<String>(10);
	static
	{	sTagsAsParam.add("space");
		sTagsAsParam.add("private");
		sTagsAsParam.add("representation");
		sTagsAsParam.add("prototype");
	}
	
	private static IPrototypeMeta lookupTag(String aTag)
	{	return (IPrototypeMeta)sTagToPrototypeMap.get(aTag);
	}
	
	private static boolean isTagSpace(String aTag)
	{	return sTagsAsParam.contains(aTag);
	}
}
