package org.netkernel.module.standard.builtin;

import java.util.*;

import org.netkernel.container.ILogger;
import org.netkernel.container.IMessages;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.NKFContextImpl;
import org.netkernel.layer0.nkf.impl.NKFRequestImpl;
import org.netkernel.layer0.representation.WrappedThrowable;
import org.netkernel.layer0.util.RequestBuilder;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.RequestBuilder.Arguments;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;
import org.netkernel.util.Utils;
import org.w3c.dom.Document;

public class PluggableOverlay extends ConfiguredOverlayImpl
{
	public PluggableOverlay()
	{	declareThreadSafe();
		declareParameters(PluggableOverlayPrototype.sPrototypeParameters);
		declareUnhandledExceptionsExpired(false);

	}
	
	@Override
	protected IConfig createConfig(INKFRequestContext aContext) throws Exception
	{	IMessages messages=getKernel().getLogger();
	
		RequestBuilder postRequestBuilder=null;
		if (aContext.exists("param:postProcess"))
		{	Document d=aContext.source("param:postProcess",Document.class);
			postRequestBuilder=new RequestBuilder(d,messages);
		}
		RequestBuilder preRequestBuilder=null;
		if (aContext.exists("param:preProcess"))
		{	Document d=aContext.source("param:preProcess",Document.class);
			preRequestBuilder=new RequestBuilder(d,messages);
		}
		RequestBuilder exRequestBuilder=null;
		if (aContext.exists("param:exceptionProcess"))
		{	Document d=aContext.source("param:exceptionProcess",Document.class);
			exRequestBuilder=new RequestBuilder(d,messages);
		}
		
		return new PluggableConfig(preRequestBuilder,postRequestBuilder,exRequestBuilder,getDelegateSpace(),getDelegateMeta(aContext),aContext,this);
	}

	@Override
	protected void destroyConfig(IConfig config) throws Exception
	{
	}

	private static class PluggableConfig implements IConfig
	{
		private ISpace mDelegateSpace;
		private ISpaceElements mElements;
		private Map<String, IMetaRepresentation> mElementMetasById;
		private RequestBuilder mPreProcess;
		private RequestBuilder mPostProcess;
		private RequestBuilder mExProcess;
		
		public PluggableConfig(RequestBuilder aPreProcess, RequestBuilder aPostProcess, RequestBuilder aExProcess, ISpace aDelegateSpace, ISpaceMeta aDelegateMeta, INKFRequestContext aContext, IEndpoint aEndpoint) throws Exception
		{
			mDelegateSpace=aDelegateSpace;
			
			mPreProcess=aPreProcess;
			mPostProcess=aPostProcess;
			mExProcess=aExProcess;
			
			//construct space elements, and get all elements meta
			ISpaceElements delegateElements=aDelegateMeta.getElements();			
			mElementMetasById = new LinkedHashMap<String, IMetaRepresentation>();
			List<String> unreferencedTargets = new ArrayList<String>();
			for (int i=0; i<delegateElements.size(); i++)
			{	IIdentifier id=delegateElements.getIdentifier(i);
				if (id!=null)
				{	String elementId = id.toString();
					unreferencedTargets.add(elementId);
					IMetaRepresentation meta=aContext.meta(elementId);
					mElementMetasById.put(elementId, meta);
				}
			}
			// build our list of elements
			TupleList tl=new TupleList(3,delegateElements.size());
			for (int i=0; i<delegateElements.size(); i++)
			{	tl.put(delegateElements.getIdentifier(i),aDelegateSpace,aEndpoint);
			}
			mElements=new SpaceElementsImpl(tl);			
		}
		
		public IMetaRepresentation getElementMeta(String elementId)
		{	return mElementMetasById.get(elementId);
		}

		public ISpaceElements getElements(INKFRequestContext aContext) throws Exception
		{	return mElements;
		}

		public String getResolvedElement(INKFRequestReadOnly aToResolve) throws Exception
		{	String result=null;
			for (Map.Entry<String, IMetaRepresentation> entry: mElementMetasById.entrySet())
			{	IMetaRepresentation meta=entry.getValue();
				if (meta instanceof IEndpointMeta)
				{	if (MetadataUtils.match((IEndpointMeta)meta, aToResolve))
					{	result = entry.getKey();
						break;
					}
				}
			}
			return result;
		}

		public void onRequest(String elementId, INKFRequestContext aContext) throws Exception
		{
			INKFRequestReadOnly reqOut=aContext.getThisRequest();
			INKFRequest reqIn;
			
			if (mPreProcess!=null)
			{	Arguments args = new Arguments();
				args.addArgumentByValue("arg:request",reqOut);
				INKFRequest preReq=mPreProcess.buildRequest(aContext, args,null);
				preReq.injectRequestScope(mDelegateSpace);
				preReq.setRepresentationClass(INKFRequest.class);
				INKFResponseReadOnly resp=aContext.issueRequestForResponse(preReq);
				Object preRepresentation = resp.getRepresentation();
				if (preRepresentation instanceof NKFRequestImpl)
				{	
					NKFRequestImpl req=(NKFRequestImpl)preRepresentation;
					req.setContext((NKFContextImpl)aContext); //ensure parent will be inherited correctly
					reqIn=req;
				}
				else
				{	ILogger logger=aContext.getKernelContext().getKernel().getLogger();
					throw Utils.createFormattedException("EX_STD_PLUGGABLE_PRE_RESP", "MSG_STD_PLUGGABLE_PRE_RESP", logger, preRepresentation.getClass().getName());
				}
			}
			else
			{	reqIn=reqOut.getIssuableClone();
			}
			
			reqIn.injectRequestScope(mDelegateSpace);
			try
			{	INKFResponseReadOnly resp=aContext.issueRequestForResponse(reqIn);

				if (mPostProcess!=null)
				{	Arguments args = new Arguments();
					args.addArgumentFromResponse("arg:response",resp);
					args.addArgumentByValue("arg:request",reqOut);
					INKFRequest postReq=mPostProcess.buildRequest(aContext, args, null);
					postReq.injectRequestScope(mDelegateSpace);
					postReq.setRepresentationClass(reqOut.getRepresentationClass());
					resp=aContext.issueRequestForResponse(postReq);
				}
				aContext.createResponseFrom(resp);
			}
			catch (NKFException e)
			{
				if (mExProcess!=null)
				{	Arguments args = new Arguments();
					args.addArgumentByValue("arg:exception",new WrappedThrowable(e));
					args.addArgumentByValue("arg:request",reqOut);
					INKFRequest exReq=mExProcess.buildRequest(aContext, args, null);
					exReq.injectRequestScope(mDelegateSpace);
					exReq.setRepresentationClass(reqOut.getRepresentationClass());
					INKFResponseReadOnly resp=aContext.issueRequestForResponse(exReq);
					aContext.createResponseFrom(resp);
				}
				else
				{	throw e;
				}
			}
		}

		public ISpace[] getSpaces(INKFRequestContext context) throws Exception
		{	return IStandardEndpoint.NO_SPACES;
		}	
	}
}
