package org.netkernel.module.standard.endpoint;

import org.netkernel.container.IKernel;
import org.netkernel.layer0.meta.GrammarException;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.PrototypeParameterMetaImpl;
import org.netkernel.layer0.meta.impl.SimpleSpaceElementsImpl;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.nkf.NKFException;
import org.netkernel.layer0.nkf.impl.NKFGrammarRequestReadOnlyImpl;
import org.netkernel.layer0.nkf.impl.NKFRequestReadOnlyImpl;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.module.standard.StandardSpace;
import org.netkernel.request.IRequest;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.ISpace;


/** Endpoint implementation that exposes a single static logical endpoint
 * @author Tony Butterfield
 */
public abstract class StandardMonoEndpointImpl extends StandardEndpointImpl
{
	protected IEndpointMeta mMeta;
	protected String mId;

	public static final String PARAM_ID="id";
	public static final String PARAM_DOC="doc";
	public static final String PARAM_ICON="icon";
	public static final String PARAM_NAME="name";
	public static final String PARAM_DESCRIPTION="description";
	public static final String PARAM_PRIVATE="private";
	
	public static IPrototypeParameterMeta[] sPrototypeParameters = new IPrototypeParameterMeta[]
	{	new PrototypeParameterMetaImpl(PARAM_ID,String.class,null),
		new PrototypeParameterMetaImpl(PARAM_NAME,"Name of endpoint",String.class,null),
		new PrototypeParameterMetaImpl(PARAM_DESCRIPTION,"Description of endpoint",String.class,null),
		new PrototypeParameterMetaImpl(PARAM_DOC, "Documentation resource identifier", IIdentifier.class,null),
		new PrototypeParameterMetaImpl(PARAM_ICON, "Endpoint icon identifier", IIdentifier.class,null),
		new PrototypeParameterMetaImpl(PARAM_PRIVATE, Boolean.class,false)
	};
	
	public StandardMonoEndpointImpl()
	{	declareParameters(sPrototypeParameters);
	}
	
	public void onCommissionEndpoint(IKernel aKernel, ISpace aSpace) throws Exception
	{	super.onCommissionEndpoint(aKernel, aSpace);

		
		String id=null;
		if (getParameters().paramExists(PARAM_ID))
		{	id=(String)getParameters().getParamValue(PARAM_ID);
		}
		if(id!=null)
		{	mId=id;
		}
		else
		{	String index;
			if (aSpace instanceof StandardSpace)
			{	index=Integer.toString(getStandardSpace().getIndexOfMapping(this));
			}
			else
			{	index=null;
			}
			mId = Utils.generateUniqueEndpointIdentifier(aSpace, this, index);
		}	
	}
	
	public ISpaceElements getParts(INKFRequestContext aContext) throws NKFException
	{	if (mId!=null)
		{	return new SimpleSpaceElementsImpl(new SimpleIdentifierImpl(mId),this);
		}
		else
		{	return SpaceElementsImpl.EMPTY;
		}
	}
	
	public ISpace[] getSpaces(INKFRequestContext context) throws Exception
	{	return IStandardEndpoint.NO_SPACES;
	}	

	public void onResolve(INKFResolutionContext aContext) throws Exception
	{	super.onResolve(aContext);
		INKFRequestReadOnly req=aContext.getRequestToResolve();
		if (mMeta!=null)
		{	if (MetadataUtils.match(mMeta, req) && additionalResolveChecks(aContext))
			{	aContext.createResolutionResponse(this,mId);
			}
		}
		else
		{	aContext.logFormatted(INKFRequestContext.LEVEL_WARNING, "MSG_STD_ENDPOINT_NO_META", new Object[]{this});
		}
	}
	
	/** Override this method to provide additional resolve checks over and beyond the basic grammar and verb checking
	 * provided by the interface metadata definition
	 * @param aContext
	 * @return true if this endpoint will resolve the request
	 * @throws Exception
	 */
	protected boolean additionalResolveChecks(INKFResolutionContext aContext) throws Exception
	{	return true;
	}
	
	public void onMeta(INKFRequestContext aContext) throws Exception
	{	aContext.createResponseFrom(mMeta);
	}
	
	protected NKFRequestReadOnlyImpl createThisNKFRequest(IRequest aRequest) throws GrammarException
	{	NKFRequestReadOnlyImpl result;
		if (mMeta==null || mMeta.getIdentifierGrammar()==null)
		{	result = new NKFRequestReadOnlyImpl(aRequest);
		}
		else
		{	result = new NKFGrammarRequestReadOnlyImpl(aRequest,mMeta.getIdentifierGrammar());
		}
		return result;
	}		

	/*Temporary hack to expose meta to subclasses*/
	protected IEndpointMeta getMeta()
	{	return mMeta;		
	}

	
}
