package org.netkernel.module.standard.endpoint;

import java.util.*;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IPrototypeParameterMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.MetadataUtils;
import org.netkernel.layer0.meta.impl.PrototypeParameterMetaImpl;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFRequestReadOnly;
import org.netkernel.layer0.nkf.INKFResolutionContext;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.ISpace;

/** Base class for a transparent overlay without configuration. Transparent overlays transmit all
 *  resources which their containing space emits. However they get to intercept the request
 *  and relay the response. If an overlay needs some kind of configuration than is dynamic and/or
 *  involves issuing requests into the space to retrieve resource then should use the base class
 *  TransparentConfiguredOverlayImpl. If an overlay presents a different set of resources that the
 *  containing space then it should extend the base class OverlayImpl
 *  
 * @see TransparentConfiguredOverlayImpl
 * @author tab
 *
 */
public abstract class TransparentOverlayImpl extends OverlayImpl
{
	protected Map<String, IMetaRepresentation> mElementMetasById;
	private boolean mBinding=false;
	
	public static IPrototypeParameterMeta[] sPrototypeParameters = new IPrototypeParameterMeta[]
    {	new PrototypeParameterMetaImpl(IStandardOverlay.PARAM_SPACE,ISpace.class,IPrototypeParameterMeta.NO_DEFAULT)
	};
	
	public TransparentOverlayImpl()
	{	declareParameters(sPrototypeParameters);
	}
	
	public ISpaceElements getParts(INKFRequestContext aContext) throws Exception
	{	ISpaceMeta delegateMeta = getDelegateMeta(aContext);
		ISpace delegateSpace= getDelegateSpace();
		IRequestScopeLevel scope = RequestScopeLevelImpl.createOrphanedRootScopeLevel(delegateSpace,aContext.getKernelContext().getRequestScope());
		aContext.getKernelContext().setRequestScope(scope);
	
		//construct space elements, and get all elements meta
		ISpaceElements delegateElements=delegateMeta.getElements();	
		LinkedHashMap<String, IMetaRepresentation> ebyid = new LinkedHashMap<String, IMetaRepresentation>();
		for (int i=0; i<delegateElements.size(); i++)
		{	IIdentifier id=delegateElements.getIdentifier(i);
			if (id!=null)
			{	String elementId = id.toString();
				IMetaRepresentation meta=aContext.meta(elementId);
				ebyid.put(elementId, meta);
			}
		}
		// build our list of elements
		TupleList tl=new TupleList(3,delegateElements.size());
		for (int i=0; i<delegateElements.size(); i++)
		{	tl.put(delegateElements.getIdentifier(i),delegateSpace,this);
		}
		mElementMetasById=ebyid;
		return new SpaceElementsImpl(tl);
	}
	
	public void onResolve(INKFResolutionContext aContext) throws Exception
	{	INKFRequestReadOnly toResolve = aContext.getRequestToResolve();
		Object requestorOverlay = toResolve.getHeaderValue(HEADER_REQUESTOR_OVERLAY);
		if (requestorOverlay==this)
		{	//overlay initiated request is trying to resolve outside overlay
			aContext.createResolutionTerminateResponse();
		}
		else
		{	if (mElementMetasById==null)
			{	if(mBinding)
				{	aContext.createResolutionTerminateResponse();
					return;
				}
				else
				{	mBinding=true;
					try
					{	getParts(aContext);
					}
					finally
					{	mBinding=false;
					}
				}
			}
			
			for (Map.Entry<String, IMetaRepresentation> entry: mElementMetasById.entrySet())
			{	IMetaRepresentation meta=entry.getValue();
				if (meta instanceof IEndpointMeta)
				{	if (MetadataUtils.match((IEndpointMeta)meta, toResolve))
					{	String endpointId = entry.getKey();
						aContext.createResolutionResponse(this, endpointId);
						break;
					}
				}
			}
		}
	}	
	
	public void onRequest(INKFRequestContext aContext) throws Exception
	{	INKFRequestReadOnly req = aContext.getThisRequest();
		if (req.getVerb()==INKFRequestReadOnly.VERB_META)
		{	Utils.delegateRequestInto(getDelegateSpace(), aContext);
		}
		else
		{	String id=aContext.getThisRequest().getResolvedElementId();
			IRequestScopeLevel scope=aContext.getKernelContext().getRequestScope();
			scope=RequestScopeLevelImpl.createNonDurableScopeLevel(getDelegateSpace(), scope);
			aContext.getKernelContext().setRequestScope(scope);	
			onRequest(id,aContext);
		}
	}
	
	/** Implement this method to intercept, request, relay on to inner space
	 * and return response.
	 * @param elementId the resolved logical endpoint from the contained space
	 * @param aContext the context for processing this request
	 * @throws Exception may be thrown for any reason you need
	 */
	public abstract void onRequest(String elementId, INKFRequestContext aContext) throws Exception;

}
